import java.util.ArrayList;

public class LowerEnvelope {

	public LowerEnvelope(Trajectory[] trSet, int b, int e, double tb, double te){
		Interval[] in=LE_Alg(trSet,b,e,tb,te);
//		for (int i=0;i<in.length;i++)
//			System.out.println(in[i].toString());
		
		LENV=in;
	}
	public Interval[] LENV;
	
	public void printEnvelope(){
		for (int i=0;i<LENV.length;i++)
			System.out.println(LENV[i].toString());
	}
	
	public Interval[] LE_Alg(Trajectory[] STR, int b, int e, double tb, double te){
		//******Algorithm 1
		/*INPUT :	STR->Set of trajectories/distance functions
					b->beginning index, e->end index
					tb->beginning time for entire range
					te->ending time for entire range
		OUTPUT :	
		*/
		if (b==e){//base case
			Interval[] base=new Interval[1];
			base[0]=new Interval(STR[b], tb, te);
			return base;
		}
		int c=(e-b)/2;
		return Merge_LE(LE_Alg(STR, b, b+c, tb, te), LE_Alg(STR, b+c+1, e, tb, te));
	}
	private Interval[] Merge_LE(Interval[] LE1, Interval[] LE2){
		/*
		 * INPUT :	LE1, LE2 - Two obtained lower envelopes we wish to merge
		 * OUTPUT :	LE12 - The combined lower envelope
		*/
		int k=0, p=0, m=LE1.length, n=LE2.length;
		Interval[] LE12=new Interval[0];
		while ((k<m)||(p<n)){
			double tcl1=LE1[k].getStart();
			double tcl2=LE2[p].getStart();
			double tcu1=LE1[k].getEnd();
			double tcu2=LE2[p].getEnd();
			
			double tcl=(tcl1>tcl2)? tcl1: tcl2;//max of tcl1,tcl2 - current lower bound
			double tcu=(tcu1<tcu2)? tcu1: tcu2;//min of tcu1, tcu2 - current upper bound
			Trajectory TR1=LE1[k].getTrajectory();
			//System.out.println("TR1 : "+TR1);
			Trajectory TR2=LE2[p].getTrajectory();
			//System.out.println("TR2 : "+TR2);
			Interval[] curEnvelope=Env2(TR1, TR2, tcl, tcu);
			
			//perform concatenation here ...
			LE12=concatenate(LE12, curEnvelope);
			
			if (tcu1<tcu2)k++;
			else if (tcu2<tcu1)p++;
			else{//tcu1==tcu2
				p++;
				k++;
			}
		}
		return LE12;
	}
	private Interval[] Env2(Trajectory TR1, Trajectory TR2, double t1, double t2){
		//used by Merge_LE
		/*
		INPUT:	TR1, TR2 - current trajectories from Merge_LE
				t1...t2 - time interval
		OUTPUT: LE[] - combined lower envelope of TR1, TR2 over tb...te
		*/
		Interval[] LE;
		double a=TR1.getA()-TR2.getA(), b=TR1.getB()-TR2.getB(), c=TR1.getC()-TR2.getC();
		if (intersect(a,b,c)){//only do calculations if parabolas intersect
			double[] criticals;
			criticals=criticalsInRange(a,b,c,t1,t2);
			LE=new Interval[criticals.length-1];
			for (int i=0;i<criticals.length-1;i++){//iterate through each critical point
				double ts=criticals[i];
				double te=criticals[i+1];
				double test_point=(ts+te)/2.0;
				double d1=evaluateDistance(TR1.getA(),TR1.getB(),TR1.getC(),test_point);
				double d2=evaluateDistance(TR2.getA(),TR2.getB(),TR2.getC(),test_point);
				if (d1<d2){//TR1 is closer trajectory during ts...te
					LE[i]=new Interval(TR1, ts, te);//append interval i to result
				}
				else{//TR2 is closer during tb...te
					LE[i]=new Interval(TR2, ts, te);
				}				
			}//end for c...
		}
		else{//do stuff here
			//****PUT STUFF HERE FOR NON-INTERSECTION****\\
			LE=new Interval[1];
			double test_point=(t1+t2)/2.0;
			if (evaluateDistance(TR1.getA(),TR1.getB(),TR1.getC(),test_point)<
				evaluateDistance(TR2.getA(),TR2.getB(),TR2.getC(),test_point)){
				LE[0]=new Interval(TR1, t1, t2);
			}
			else LE[0]=new Interval(TR2, t1, t2);
		}
		//for (int i=0;i<LE.length;i++)System.out.println(LE[i]);
		return LE;
	}
	
	private Interval[] concatenate(Interval[] LE12, Interval[] env2){
		ArrayList<Interval> ret=new ArrayList<Interval>();
		for (int i=0;i<LE12.length;i++){
			ret.add(LE12[i]);
		}
		int lastInd=ret.size()-1;
		for (int i=0;i<env2.length;i++){
			//check for same string...
			if ((LE12.length==0)||(!env2[i].getTrajectory().
					getId().equals(LE12[lastInd].getTrajectory().getId()))){
				//System.out.println(env2[i]);
				ret.add(env2[i]);
			}
			else{
				//traj to add is the same as previous, so modify
				//last element of LE12
				
				double tb=LE12[lastInd].getStart();
				double te=env2[i].getEnd();
				ret.remove(ret.size()-1);
				ret.add(new Interval(LE12[lastInd].getTrajectory(),tb,te));
			}
		}
		//set up return value
		Interval[] conc=new Interval[ret.size()];
		for (int i=0;i<conc.length;i++){
			//System.out.println((Interval)ret.get(i));
			conc[i]=new Interval((Interval)ret.get(i));
		}
		return conc;
		//return (Interval[])ret.toArray();
	}

	private double[] criticalsInRange(double a, double b, double c, double tb, double te){
		/*
		 * INPUT :	a,b,c - floats specifying parabola
		 * tb, te - range we're considering
		 * 
		 * OUPUT :		criticals - All critical points, ordered in the range tb...te.
		 *  These are tb, te, and the 0-points for parabola a,b,c that
		 *  are between tb...te
		 */
		double[] roots=findRoots(a, b, c);
		double[] criticals;
		int critCount=0;
		//first count# of critical points
		for (int i=0;i<roots.length;i++)
			if (roots[i]>tb&&roots[i]<te)//point from intersection is in range
				critCount++;
		//now set values of each critical
		criticals=new double[critCount+2];
		int index=1;
		criticals[0]=tb;
		for (int i=0;i<roots.length;i++)
			if (roots[i]>tb&&roots[i]<te){
				criticals[index]=roots[i];
				index++;
			}
		criticals[index]=te;
		//for (int i=0;i<criticals.length;i++)System.out.println(criticals[i]);
		return criticals;
	}

	private double[] findRoots(double a, double b, double c){
		/*
		 * INPUT :	a,b,c - floats specifying parabola
		 * OUTPUT :	sol<> containing unique roots ordered from least to greatest
		*/
		if (a!=0){
			double d=Math.sqrt(b*b-4.0*a*c);
			double root1=(-b+d)/(2.0*a);
			double root2=(-b-d)/(2.0*a);
			double[] sol=new double[2];
			sol[0]=(root1<root2) ? root1 : root2;
			sol[1]=(root1<root2) ? root2: root1;
			return sol;
		}
		else{
			//line is of the form y=bx+c, solution is 0=bx+c or x=-c/b
			double[] sol=new double[1];
			sol[0]=-c/b;
			return sol;
		}
	}

	private boolean intersect(double a, double b, double c){
		//Does the intersection parabola ax^2+bx+c (a1-a2,b1-b2,c1-c2) cross the x-axis?
		//if not, there is no intersection
		return ((b*b-4.0*a*c)>0);
	}
	
	private double evaluateDistance(double a, double b, double c, double t){
		//evaluates parabola ax^2+bx+c at x=t
		return a*t*t+b*t+c;
	}
}
